class System():
  def __init__(self, panel, label, status):
      """Read system from response."""
      self._panel = panel
      self._status = status
      self._name = label.strip()

  @property
  def name(self):
      """System name."""
      return self._name

  @property
  def low_battery_trouble(self):
      return 'B' in self._status

  @property
  def ac_trouble(self):
      return 'A' in self._status

  @property
  def phone_line_trouble(self):
      return 'P' in self._status

  @property
  def clock_trouble(self):
      return 'C' in self._status

  @property
  def box_tamper(self):
      return 'X' in self._status

  @property
  def programming_mode(self):
      return 'I' in self._status

  def update_status(self, status):
    self._status = status



import asyncio
from .risco_crypt import RiscoCrypt, ESCAPED_END, END
from pyrisco.common import UnauthorizedError, CannotConnectError, OperationError

MIN_CMD_ID = 1
MAX_CMD_ID = 49

class RiscoSocket:
  def __init__(self, host, port, code, **kwargs):
    self._host = host
    self._port = port
    self._code = code
    self._encoding = kwargs.get('encoding', 'utf-8')
    self._max_concurrency = kwargs.get('concurrency', 4)
    self._communication_delay = kwargs.get('communication_delay', 0)
    self._reader = None
    self._writer = None
    self._crypt = None
    self._listen_task = None
    self._keep_alive_task = None
    self._semaphore = None
    self._queue = None

  @property
  def queue(self):
    return self._queue

  async def connect(self):
    self._cmd_id = 0
    try:
      self._semaphore = asyncio.Semaphore(self._max_concurrency)
      self._futures = [None for i in range(MIN_CMD_ID, MIN_CMD_ID + MAX_CMD_ID)]
      self._reader, self._writer = await asyncio.open_connection(self._host, self._port)
      if self._communication_delay > 0:
        await asyncio.sleep(self._communication_delay)
      self._queue = asyncio.Queue()
      self._listen_task = asyncio.create_task(self._listen())
      self._crypt = RiscoCrypt(self._encoding)
      panel_id = int(await self.send_result_command('RID'), 16)
      self._crypt.set_panel_id(panel_id)
      if not await self.send_ack_command('LCL'):
        raise CannotConnectError
      command = f'RMT={self._code}'
      if not await self.send_ack_command(command):
        raise UnauthorizedError

      self._keep_alive_task = asyncio.create_task(self._keep_alive())
    except Exception as exc:
      await self._close()
      raise CannotConnectError from exc

  async def disconnect(self):
    if self._writer:
      try:
        await self.send_ack_command('DCN')
      except OperationError:
        # safe to ignore these when disconnecting
        pass
      finally:
        await self._close()

  async def _listen(self):
    while True:
      try:
        cmd_id, command, crc = await self._read_command()
        if not cmd_id:
          self._decrement_cmd_id()
          raise OperationError(f'Risco error: {command}')
        if cmd_id <= MAX_CMD_ID:
          future = self._futures[cmd_id-1]
          self._futures[cmd_id-1] = None
          if not crc:
            future.set_exception(OperationError(f'cmd_id: {cmd_id}, Wrong CRC'))
          elif command[0] in ['N', 'B']:
            future.set_exception(OperationError(f'cmd_id: {cmd_id}, Risco error: {command}'))
          else:
            future.set_result(command)
        else:
          await self._handle_incoming(cmd_id, command, crc)
      except Exception as error:
        await self._queue.put(error)


  async def _keep_alive(self):
    while True:
      try:
        await self.send_result_command("CLOCK")
      except OperationError as error:
        await self._queue.put(error)

      await asyncio.sleep(5)

  async def send_ack_command(self, command):
    command = await self.send_command(command)
    return command == 'ACK'

  async def send_result_command(self, command):
    command = await self.send_command(command)
    return command.split("=")[1]

  async def send_command(self, command, force_encryption=False):
    async with self._semaphore:
      self._increment_cmd_id()
      self._write_command(self._cmd_id, command, force_encryption)
      future = asyncio.Future()
      self._futures[self._cmd_id-1] = future
      try:
        return await asyncio.wait_for(future, 1)
      except asyncio.TimeoutError:
        raise OperationError(f'Timeout in command: {command}')

  async def _handle_incoming(self, cmd_id, command, crc):
    self._write_command(cmd_id, 'ACK')
    if not crc:
      await self._queue.put(OperationError(f'cmd_id: {cmd_id}, Wrong CRC'))
    else:
      await self._queue.put(command)

  async def _read_command(self):
    buffer = await self._reader.readuntil(END)
    while buffer.endswith(ESCAPED_END):
      buffer += await self._reader.readuntil(END)
    return self._crypt.decode(buffer)

  def _write_command(self, cmd_id, command, force_encryption=False):
    buffer = self._crypt.encode(cmd_id, command, force_encryption)
    self._writer.write(buffer)

  async def _close(self):
    if self._keep_alive_task:
      self._keep_alive_task.cancel()
      self._keep_alive_task = None

    if self._listen_task:
      self._listen_task.cancel()
      self._listen_task = None

    if self._writer:
      self._writer.close()
      await self._writer.wait_closed()
    self._crypt = None
    self._writer = None
    self._reader = None
    self._semaphore = None
    self._queue = None
    # Risco needs a few seconds to reset its encryption state before accepting a new connection
    # If we don't sleep here, the next connection will be encrypted before we get the panel id.
    await asyncio.sleep(5)

  def _increment_cmd_id(self):
    self._cmd_id += 1
    if self._cmd_id > MAX_CMD_ID:
      self._cmd_id = MIN_CMD_ID

  def _decrement_cmd_id(self):
    self._cmd_id -= 1
    if self._cmd_id < MIN_CMD_ID:
      self._cmd_id = MAX_CMD_ID



"""Implementation of a Risco Cloud connection."""

import aiohttp
import asyncio

from .alarm import Alarm
from .event import Event
from pyrisco.common import UnauthorizedError, CannotConnectError, OperationError, GROUP_ID_TO_NAME


LOGIN_URL = "https://www.riscocloud.com/webapi/api/auth/login"
SITE_URL = "https://www.riscocloud.com/webapi/api/wuws/site/GetAll"
PIN_URL = "https://www.riscocloud.com/webapi/api/wuws/site/%s/Login"
STATE_URL = "https://www.riscocloud.com/webapi/api/wuws/site/%s/ControlPanel/GetState"
CONTROL_URL = "https://www.riscocloud.com/webapi/api/wuws/site/%s/ControlPanel/PartArm"
EVENTS_URL = (
  "https://www.riscocloud.com/webapi/api/wuws/site/%s/ControlPanel/GetEventLog"
)
BYPASS_URL = "https://www.riscocloud.com/webapi/api/wuws/site/%s/ControlPanel/SetZoneBypassStatus"

NUM_RETRIES = 3


class RiscoCloud:
  """A connection to a Risco alarm system."""

  def __init__(self, username, password, pin, language="en"):
    """Initialize the object."""
    self._username = username
    self._password = password
    self._pin = pin
    self._language = language
    self._access_token = None
    self._session_id = None
    self._site_id = None
    self._site_name = None
    self._site_uuid = None
    self._session = None
    self._created_session = False

  async def _authenticated_post(self, url, body):
    headers = {
      "Content-Type": "application/json",
      "authorization": "Bearer " + self._access_token,
    }
    async with self._session.post(url, headers=headers, json=body) as resp:
      json = await resp.json()

    if json["status"] == 401:
      raise UnauthorizedError(json["errorText"])

    if "result" in json and json["result"] != 0:
      raise OperationError(str(json))

    return json["response"]

  async def _site_post(self, url, body):
    site_url = url % self._site_id
    for i in range(NUM_RETRIES):
      try:
        site_body = {
            **body,
            "fromControlPanel": True,
            "sessionToken": self._session_id,
        }
        return await self._authenticated_post(site_url, site_body)
      except UnauthorizedError:
        if i + 1 == NUM_RETRIES:
          raise
        await self.close()
        await self.login()

  async def _login_user_pass(self):
    headers = {"Content-Type": "application/json"}
    body = {"userName": self._username, "password": self._password}
    try:
      async with self._session.post(
        LOGIN_URL, headers=headers, json=body
      ) as resp:
        json = await resp.json()
        if json["status"] == 401:
          raise UnauthorizedError("Invalid username or password")
        self._access_token = json["response"].get("accessToken")
    except aiohttp.client_exceptions.ClientConnectorError as e:
      raise CannotConnectError from e

    if not self._access_token:
      raise UnauthorizedError("Invalid username or password")

  async def _login_site(self):
    resp = await self._authenticated_post(SITE_URL, {})
    self._site_id = resp[0]["id"]
    self._site_name = resp[0]["name"]
    self._site_uuid = resp[0]["siteUUID"]

  async def _login_session(self):
    body = {"languageId": self._language, "pinCode": self._pin}
    url = PIN_URL % self._site_id
    resp = await self._authenticated_post(url, body)
    self._session_id = resp["sessionId"]

  async def _init_session(self, session):
    await self.close()
    if self._session is None:
      if session is None:
        self._session = aiohttp.ClientSession()
        self._created_session = True
      else:
        self._session = session

  async def _send_control_command(self, body):
    resp = await self._site_post(CONTROL_URL, body)
    return Alarm(self, resp)

  async def close(self):
    """Close the connection."""
    self._session_id = None
    if self._created_session == True and self._session is not None:
      await self._session.close()
      self._session = None
      self._created_session = False

  async def login(self, session=None):
    """Login to Risco Cloud."""
    if self._session_id:
        return

    await self._init_session(session)
    await self._login_user_pass()
    await self._login_site()
    await self._login_session()

  async def get_state(self):
    """Get partitions and zones."""
    resp = await self._site_post(STATE_URL, {})
    return Alarm(self, resp["state"]["status"])

  async def disarm(self, partition):
    """Disarm the alarm."""
    body = {
      "partitions": [{"id": partition, "armedState": 1}],
    }
    return await self._send_control_command(body)

  async def arm(self, partition):
    """Arm the alarm."""
    body = {
      "partitions": [{"id": partition, "armedState": 3}],
    }
    return await self._send_control_command(body)

  async def partial_arm(self, partition):
    """Partially-arm the alarm."""
    body = {
      "partitions": [{"id": partition, "armedState": 2}],
    }
    return await self._send_control_command(body)

  async def group_arm(self, partition, group):
    """Arm a specific group."""
    if isinstance(group, str):
      group = GROUP_ID_TO_NAME.index(group)

    body = {
      "partitions": [{"id": partition, "groups": [{"id": group, "state": 3}]}],
    }
    return await self._send_control_command(body)

  async def get_events(self, newer_than, count=10):
    """Get event log."""
    body = {
      "count": count,
      "newerThan": newer_than,
      "offset": 0,
    }
    response = await self._site_post(EVENTS_URL, body)
    return [Event(e) for e in response["controlPanelEventsList"]]

  async def bypass_zone(self, zone, bypass):
    """Bypass or unbypass a zone."""
    status = 2 if bypass else 3
    body = {"zones": [{"trouble": 0, "ZoneID": zone, "Status": status}]}
    resp = await self._site_post(BYPASS_URL, body)
    return Alarm(self, resp)

  @property
  def site_id(self):
    """Site ID of the Alarm instance."""
    return self._site_id

  @property
  def site_name(self):
    """Site name of the Alarm instance."""
    return self._site_name

  @property
  def site_uuid(self):
    """Site UUID of the Alarm instance."""
    return self._site_uuid



from pyrisco.common import GROUP_ID_TO_NAME

EVENT_IDS_TO_TYPES = {
  3: "triggered",
  9: "zone bypassed",
  10: "zone unbypassed",
  13: "armed",
  16: "disarmed",
  28: "power lost",
  29: "power restored",
  34: "media lost",
  35: "media restore",
  36: "service needed",
  118: "group arm",
  119: "group arm",
  120: "group arm",
  121: "group arm",
}

class Event:
  """A representation of a Risco event."""

  def __init__(self, raw):
    """Read event from response."""
    self._raw = raw

  @property
  def raw(self):
    return self._raw

  @property
  def type_id(self):
    return self.raw["eventId"]

  @property
  def type_name(self):
    return EVENT_IDS_TO_TYPES.get(self.type_id, "unknown"),

  @property
  def partition_id(self):
    partition_id = self.raw["partAssociationCSV"]
    if partition_id is None:
      return None

    return int(partition_id)

  @property
  def time(self):
    """Time the event was fired."""
    return self.raw["logTime"]

  @property
  def text(self):
    """Event text."""
    return self.raw["eventText"]

  @property
  def name(self):
    """Event name."""
    return self.raw["eventName"]

  @property
  def category_id(self):
    """Event group number."""
    return self.raw["group"]

  @property
  def category_name(self):
    """Event group number."""
    return self.raw["groupName"]

  @property
  def zone_id(self):
    if self.raw["sourceType"] == 1:
      return self.raw["sourceID"] - 1
    return None

  @property
  def user_id(self):
    if self.raw["sourceType"] == 2:
      return self.raw["sourceID"]
    return None

  @property
  def group(self):
    if self.type_id in range(118, 122):
      return GROUP_ID_TO_NAME[self.type_id - 118]
    return None

  @property
  def priority(self):
    return self.raw["priority"]

  @property
  def source_id(self):
    return self._source_id
